from abc import ABC, abstractmethod
import datetime
import hashlib
import json
import logging
import os
from pathlib import Path
import socket
import sys
from typing import Any

from .utils import safe_append_to_file, format_timedelta
from ..conversion import hash_file, create_loggers
from ..naming import ImageFile
from ..qa import create_qa_image

CONTAINER_LABELS = [
    "ci.image",
    "ci.tag",
    "ci.commit",
    "ci.builder",
    "ci.timestamp",
    "ci.digest",
]


class ProcessingModule(ABC):
    name: str = None
    version: str = None
    log_uses_filename: bool = True
    skip_prov_write: tuple[str] = tuple()

    def __init__(self, args: list[str] | None = None) -> None:
        self.verify_container()
        self.cli_call = " ".join(sys.argv[1:] if args is None else args)
        self.start_time = datetime.datetime.now()
        self.parsed_args = self.cli(args)
        if self.parsed_args is None:  # CLI call returned None, so we're done
            return
        self.create_loggers()
        try:
            logging.info(f"Beginning processing using: {self.name} v{self.version}.")
            logging.info(f"Command: {self.cli_call}")
            self.outputs = self.run(**self.parsed_args)
            if not self.check_outputs():
                logging.error(f"Processing failed. No outputs reported.")
                return
            logging.info(f"Generating QA imagees.")
            self.generate_qa_images()
            logging.info(f"Generating provenance records.")
            self.generate_prov()
            logging.info(f"Processing complete.")
        except Exception as e:
            logging.error(f"An error occurred during processing: {e}.", exc_info=True)
            raise

    @staticmethod
    @abstractmethod
    def cli(args: list[str] | None = None) -> dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def run(*args, **kwargs) -> dict[str, Path] | list[dict[str, Path]]:
        raise NotImplementedError

    @staticmethod
    def get_container_labels() -> dict[str, str]:
        with open("/.singularity.d/labels.json", "r") as f:
            labels = json.load(f)
        return labels

    @staticmethod
    def verify_container() -> None:
        if Path("/.singularity.d/labels.json").exists():
            labels = ProcessingModule.get_container_labels()
            if any(lbl not in labels for lbl in CONTAINER_LABELS):
                raise ValueError("Container is missing required labels.")
        else:
            raise ValueError("Container labels not found. Running outside of container?")

    def check_outputs(self) -> bool:
        if self.outputs is None:
            return False
        if self.check_multi_inputs():
            return (
                isinstance(self.outputs, list)
                and len(self.outputs) > 0
                and isinstance(self.outputs[0], dict)
            )
        else:
            return isinstance(self.outputs, dict) and len(self.outputs) > 0

    def create_prov(self, args: dict[str], outputs: dict[str, Path | list[Path]]) -> str:
        lbls = self.get_container_labels()
        project_root = [
            (el[0] if isinstance(el, tuple) else el)
            for sub in outputs.values()
            for el in (sub if isinstance(sub, list) else [sub])
        ][0].parent.parent.parent.parent
        user = os.environ["USER"] if "USER" in os.environ else Path(os.environ["HOME"]).name
        prov_str = (
            f"Module: {self.name}:{self.version}\n"
            f"Container: \n"
            f"  url: {lbls['ci.image']}:{lbls['ci.tag']}@{lbls['ci.commit'][:8]}\n"
            f"  hash: {lbls['ci.digest']}\n"
            f"  builder: {lbls['ci.builder']}\n"
            f"  timestamp: {lbls['ci.timestamp']}\n"
            f"User: {user}@{socket.getfqdn()}\n"
            f"StartTime: {self.start_time.isoformat(timespec='seconds')}\n"
            f"Duration: {format_timedelta(datetime.datetime.now() - self.start_time)}\n"
            f"ProjectRoot: {project_root}\n"
            f"Inputs: "
        )
        inputs = {
            k: v
            for k, v in args.items()
            if isinstance(v, (Path, ImageFile))
            or (isinstance(v, list) and isinstance(v[0], (Path, ImageFile)))
        }
        if len(inputs) > 0:
            prov_str += "\n"
            prov_str += self.get_prov_path_strs(inputs, project_root)
        else:
            prov_str += "None\n"
        prov_str += f"Outputs: \n"
        prov_str += self.get_prov_path_strs(outputs, project_root)
        params = {k: v for k, v in args.items() if k not in inputs}
        prov_str += f"Parameters: "
        if len(params) > 0:
            prov_str += "\n"
            for k, v in params.items():
                if isinstance(v, list):
                    prov_str += f"  {k}:\n"
                    for item in v:
                        prov_str += f"    - {str(item)}\n"
                else:
                    prov_str += f"  {k}: {str(v)}\n"
        else:
            prov_str += "None\n"
        prov_str += f"Command: {self.cli_call}\n"
        hashobj = hashlib.sha256()
        hashobj.update(prov_str.encode("utf-8"))
        prov_str = f"---\nId: {hashobj.hexdigest()}\n" + prov_str + f"...\n"
        return prov_str

    @staticmethod
    def get_prov_path_strs(path_dict: dict[str, Path | list[Path] | None], project_root: Path) -> str:
        prov_str = ""
        for k, v in path_dict.items():
            if v is None:
                prov_str += "f  {k}: None\n"
            elif isinstance(v, list):
                prov_str += f"  {k}:\n"
                for item in v:
                    val = item[0] if isinstance(item, tuple) else item
                    rel_path = (
                        val.relative_to(project_root) if val.is_relative_to(project_root) else val
                    )
                    prov_str += (
                        f"    - {str(rel_path)}:sha256:{hash_file(val, include_names=False)}\n"
                    )
            else:
                val = v[0] if isinstance(v, tuple) else v
                rel_path = (
                    val.relative_to(project_root) if val.is_relative_to(project_root) else val
                )
                prov_str += f"  {k}: {str(rel_path)}:sha256:{hash_file(val, include_names=False)}\n"
        return prov_str

    @staticmethod
    def write_prov(
        prov_str: str,
        outputs: dict[str, Path | list[Path] | None],
        skip_prov_write: tuple[str],
    ) -> None:
        outs = [
            (el[0] if isinstance(el, tuple) else el)
            for key, sub in outputs.items()
            if key not in skip_prov_write
            for el in (sub if isinstance(sub, list) else [sub])
        ]
        if len(outs) == 0:
            return
        first = True
        for output in outs:
            if output is None:
                continue
            if first:
                session_dir = output.parent.parent
                session_file = "_".join(
                    [session_dir.parent.name, session_dir.name, "Provenance.yml"]
                )
                safe_append_to_file(session_dir / session_file, prov_str)
                first = False
            suffix = "".join(output.suffixes)
            prov_path = output.parent / output.name.replace(suffix, ".prov")
            with open(prov_path, "w") as f:
                f.write(prov_str)

    def check_multi_inputs(self) -> bool:
        return all(
            isinstance(arg, list) and len(arg) == len(list(self.parsed_args.values())[0])
            for arg in self.parsed_args.values()
        )

    def check_multi_run(self) -> bool:
        return self.check_multi_inputs() and len(self.outputs) == len(
            list(self.parsed_args.values())[0]
        )

    def generate_prov(self) -> None:
        if self.check_multi_run():
            for i in range(len(list(self.parsed_args.values())[0])):
                prov_str = self.create_prov(
                    {k: v[i] for k, v in self.parsed_args.items()},
                    self.outputs[i],
                )
                self.write_prov(prov_str, self.outputs[i], self.skip_prov_write)
        else:
            prov_str = self.create_prov(self.parsed_args, self.outputs)
            self.write_prov(prov_str, self.outputs, self.skip_prov_write)

    @staticmethod
    def create_qa(
        outputs: dict[str, Path | list[Path] | None],
        name: str,
        skip_prov_write: tuple[str],
    ) -> None:
        outs = [
            el
            for key, sub in outputs.items()
            if key not in skip_prov_write
            for el in (sub if isinstance(sub, list) else [sub])
        ]
        if len(outs) == 0:
            return
        out = outs[0][0] if isinstance(outs[0], tuple) else outs[0]
        out_dir = out.parent.parent / "qa" / name
        out_dir.mkdir(exist_ok=True, parents=True)
        for out in outs:
            if out is None:
                continue
            if isinstance(out, tuple):
                overlay = out[0]
                bg_image = out[1]
                lut = out[2] if len(out) > 2 else "binary"
                out_name = f"{overlay.name.split('.')[0]}.png"
            else:
                overlay = None
                bg_image = out
                lut = "binary"
                out_name = f"{bg_image.name.split('.')[0]}.png"
            if not str(bg_image).endswith(".nii.gz"):
                continue
            create_qa_image(
                str(bg_image),
                out_dir / out_name,
                str(overlay) if overlay is not None else None,
                lut,
            )

    def generate_qa_images(self) -> None:
        if self.check_multi_run():
            for i in range(len(list(self.parsed_args.values())[0])):
                self.create_qa(self.outputs[i], self.name, self.skip_prov_write)
        else:
            self.create_qa(self.outputs, self.name, self.skip_prov_write)

    def create_loggers(self):
        out_paths = None
        if self.check_multi_inputs():
            for arg_list in self.parsed_args.values():
                if isinstance(arg_list[0], (Path, ImageFile)):
                    out_paths = arg_list
                    break
                elif isinstance(arg_list[0], list) and isinstance(
                    arg_list[0][0], (Path, ImageFile)
                ):
                    out_paths = [item[0] for item in arg_list]
                    break
        else:
            for arg in self.parsed_args.values():
                if isinstance(arg, (Path, ImageFile)):
                    out_paths = [arg]
                    break
        if out_paths is None:
            raise ValueError("Could not find any Path or ImageFile inputs.")
        for i, out_path in enumerate(out_paths):
            if self.log_uses_filename:
                log_dir = out_path.parent.parent / "logs" / self.name
                log_filename = out_path.name.split(".")[0]
            else:
                log_dir = out_path.parent.parent / "logs"
                log_filename = self.name
            create_loggers(log_dir, log_filename, add_stream_handler=i == 0)



from __future__ import annotations

from functools import cached_property
from glob import iglob as standard_iglob
import json
import os
from pathlib import Path
from typing import Generator


class ImageFile:
    def __init__(self, path: str | os.PathLike[str]) -> None:
        self._path = Path(path).resolve()
        self._info = None

    def open(self, mode: str = "r"):
        return self.path.open(mode)

    def __str__(self):
        return str(self.path)

    @property
    def info(self) -> ImageInfo:
        if self._info is None:
            if self.path.parent.name == "nii":
                self._info = ImageInfo(self.path.parent / self.path.name.replace(self.ext, ".json"))
            elif self.path.parent.name == "proc" or self.path.parent.name == "stage":
                self._info = ImageInfo(
                    self.path.parent.parent / "nii" / ("_".join(self._name_arr[:4]) + ".json")
                )
            else:
                raise ValueError(
                    f"ImageFile.info is not available for {self.path}. "
                    f"Check the parent folders."
                )
        return self._info

    @property
    def path(self) -> Path:
        return self._path

    def is_relative_to(self, other: Path | str) -> bool:
        return self.path.is_relative_to(other)

    def relative_to(self, other: Path | str) -> Path:
        return self.path.relative_to(other)

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def stem(self) -> str:
        return self.path.name.split(".")[0]

    @property
    def suffixes(self) -> list[str]:
        return self.path.suffixes

    @property
    def ext(self) -> str:
        return "".join(self.suffixes)

    @cached_property
    def _name_arr(self) -> list[str]:
        return self.name.split(".")[0].split("_")

    @property
    def parent(self) -> Path:
        return self.path.parent

    @property
    def subject_id(self) -> str:
        return self._name_arr[0]

    @property
    def session_id(self) -> str:
        return self._name_arr[1]

    @property
    def image_id(self) -> str:
        return self._name_arr[2]

    @property
    def series_id(self) -> str:
        return self.image_id.split("-")[0]

    @property
    def image_type(self) -> str:
        return self._name_arr[3]

    @cached_property
    def _image_type_arr(self) -> list[str]:
        return self._name_arr[3].split("-")

    @property
    def bodypart(self) -> str:
        return self._image_type_arr[0]

    @property
    def modality(self) -> str:
        return self._image_type_arr[1]

    @property
    def technique(self) -> str:
        return self._image_type_arr[2]

    @property
    def acqdim(self) -> str:
        return self._image_type_arr[3]

    @property
    def orientation(self) -> str:
        return self._image_type_arr[4]

    @property
    def excontrast(self) -> str:
        return self._image_type_arr[5]

    @property
    def extras(self) -> list[str]:
        if len(self._image_type_arr) < 7:
            return []
        return self._image_type_arr[6:]

    @property
    def tags(self) -> list[str]:
        if len(self._name_arr) < 5:
            return []
        return self._name_arr[4:]


class ImageInfo:
    def __init__(self, path: str | os.PathLike[str]) -> None:
        self._path = Path(path).resolve()
        self._info = json.loads(self._path.read_text())["SeriesInfo"]

    def __getattr__(self, item):
        if item in self._info:
            return self._info[item]
        item = "".join([i.title() for i in item.split("_")])
        if item in self._info:
            return self._info[item]
        raise AttributeError(f"No attribute {item} found.")


class ImageFilter:
    allowed_keys = (
        "bodypart",
        "modality",
        "technique",
        "acqdim",
        "orientation",
        "excontrast",
        "extras",
        "tags",
    )

    def __init__(self, **kwargs):
        for key in kwargs.keys():
            if key not in self.allowed_keys:
                raise ValueError(
                    f"Invalid key provided {key}. Allowed keys are: {', '.join(self.allowed_keys)}\n"
                    f"Keys provided: {', '.join(kwargs.keys())}"
                )
        self._filter_dict = kwargs

    def __str__(self):
        return ";".join([f"{key}={str(value)}" for key, value in self._filter_dict.items()])

    def __getattr__(self, item):
        if item in self.allowed_keys:
            return self._filter_dict.get(item)
        raise AttributeError(
            f"Invalid key provided {item}. "
            f"Allowed keys are: {', '.join(self.allowed_keys)}"
        )

    @classmethod
    def from_string(cls, filter_str) -> ImageFilter:
        filter_dict = dict([item.split("=") for item in filter_str.split(";")])
        for key, value in filter_dict.items():
            if key in ["extras", "tags"]:
                if value.upper() in ["NONE", "()", "[]"]:
                    filter_dict[key] = []
                else:
                    if value[0] in ["(", "["]:
                        value = value[1:]
                    if value[-1] in [")", "]"]:
                        value = value[:-1]
                    if key == "extras":
                        filter_dict[key] = [item.upper() for item in value.split(",")]
                    else:
                        filter_dict[key] = value.split(",")
            else:
                filter_dict[key] = value.upper()
        return cls(**filter_dict)

    def iterfilter(self, imgs: list[ImageFile]) -> Generator[ImageFile, None, None]:
        for img in imgs:
            if self.check(img):
                yield img

    def filter(self, imgs: list[ImageFile]) -> list[ImageFile]:
        return list(self.iterfilter(imgs))

    def check(self, img: ImageFile) -> bool:
        return all(
            [self.match_attr(getattr(img, key), value) for key, value in self._filter_dict.items()]
        )

    @staticmethod
    def match_attr(img_value, dict_value) -> bool:
        if isinstance(dict_value, list):
            return len(img_value) == len(dict_value) and all(
                x == y for x, y in zip(img_value, dict_value)
            )
        elif callable(dict_value):
            return dict_value(img_value)
        return img_value == dict_value


def iglob(path: str | os.PathLike[str], recursive: bool = False) -> list[ImageFile]:
    for p in standard_iglob(str(path), recursive=recursive):
        yield ImageFile(p)


def glob(path: str | os.PathLike[str], recursive: bool = False) -> list[ImageFile]:
    return list(iglob(path, recursive=recursive))



import json
from pathlib import PurePath
from typing import Any, Union
import uuid


class NoIndent:
    def __init__(self, value: Any) -> None:
        self.value = value


class JSONObjectEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.kwargs = dict(kwargs)
        del self.kwargs["indent"]
        self._replacement_map = {}

    def default(self, o: Any) -> Union[str, dict]:
        if isinstance(o, NoIndent):
            key = uuid.uuid4().hex
            self._replacement_map[key] = json.dumps(o.value, **self.kwargs, cls=JSONObjectEncoder)
            return "@@%s@@" % (key,)
        elif isinstance(o, PurePath):
            return str(o)
        elif hasattr(o, "__repr_json__") and callable(o.__repr_json__):
            return o.__repr_json__()
        else:
            return super().default(o)

    def encode(self, o: Any) -> str:
        result = super().encode(o)
        for k, v in self._replacement_map.items():
            result = result.replace('"@@%s@@"' % (k,), v)
        return result



import argparse
import json
import logging
from pathlib import Path
import shutil
from typing import List, Optional

from .._version import __version__
from .exec import run_conversion, ExecError
from .lut import LookupTable
from .metadata import Metadata
from .utils import hash_file_dir, silentremove, mkdir_p, version_check


def convert(args: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("source", type=Path, help="Source directory/file to convert.")
    parser.add_argument(
        "-o", "--output-root", type=Path, help="Output root directory.", required=True
    )
    parser.add_argument("-l", "--lut-file", type=Path, help="Lookup table file.")
    parser.add_argument("-p", "--project-id", type=str, help="Project ID.")
    parser.add_argument("-s", "--subject-id", type=str, help="Subject ID.")
    parser.add_argument("-e", "--session-id", type=str, help="Session ID.")
    parser.add_argument("--site-id", type=str, help="Site ID.")
    parser.add_argument("--tms-metafile", type=Path, help="TMS metadata file.")
    parser.add_argument("--verbose", action="store_true", help="Verbose output.")
    parser.add_argument(
        "--force", action="store_true", help="Force run even if it would be skipped."
    )
    parser.add_argument(
        "--reckless", action="store_true", help="Force run and overwrite existing data."
    )
    parser.add_argument(
        "--safe", action="store_true", help="Add -N to session ID, if session exists."
    )
    parser.add_argument(
        "--no-project-subdir", action="store_true", help="Do not create project subdirectory."
    )
    parser.add_argument("--parrec", action="store_true", help="Source is PARREC.")
    parser.add_argument(
        "--symlink",
        action="store_true",
        help="Create symbolic links to source data instead of copying.",
    )
    parser.add_argument(
        "--hardlink",
        action="store_true",
        help="Create hard links to source data instead of copying.",
    )
    parser.add_argument("--institution", type=str, help="Institution name.")
    parser.add_argument("--field-strength", type=int, help="Magnetic field strength.")
    parser.add_argument(
        "--force-dicom", action="store_true", help="Force read DICOM files.", default=False
    )
    parser.add_argument("--anonymize", action="store_true", help="Anonymize DICOM data.")
    parser.add_argument("--date-shift-days", type=int, help="Number of days to shift dates.")
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

    args = parser.parse_args(args)

    for argname in ["source", "output_root", "lut_file", "tms_metafile"]:
        if getattr(args, argname) is not None:
            setattr(args, argname, getattr(args, argname).resolve())

    if args.hardlink and args.symlink:
        raise ValueError("Only one of --symlink and --hardlink can be used.")
    linking = "hardlink" if args.hardlink else ("symlink" if args.symlink else None)

    mapping = {"subject_id": "SubjectID", "session_id": "SessionID", "site_id": "SiteID"}
    if args.tms_metafile:
        metadata = Metadata.from_tms_metadata(args.tms_metafile, args.no_project_subdir)
        for argname in ["subject_id", "session_id", "site_id"]:
            if getattr(args, argname) is not None:
                setattr(metadata, mapping[argname], getattr(args, argname))
    else:
        for argname in ["project_id", "subject_id", "session_id"]:
            if getattr(args, argname) is None:
                raise ValueError(
                    "%s is a required argument when no metadata file is provided." % argname
                )
        metadata = Metadata(
            args.project_id,
            args.subject_id,
            args.session_id,
            args.site_id,
            args.no_project_subdir,
        )

    if args.lut_file is None:
        lut_file = (
            (args.output_root / (metadata.projectname + "-lut.csv"))
            if args.no_project_subdir
            else (args.output_root / metadata.projectname / (metadata.projectname + "-lut.csv"))
        )
    else:
        lut_file = args.lut_file

    manual_json_file = (
        args.output_root / metadata.dir_to_str() / (metadata.prefix_to_str() + "_ManualNaming.json")
    )
    manual_names = json.loads(manual_json_file.read_text()) if manual_json_file.exists() else {}

    type_dirname = "%s" % "parrec" if args.parrec else "dcm"
    if (args.output_root / metadata.dir_to_str() / type_dirname).exists():
        if args.safe:
            metadata.AttemptNum = 2
            while (args.output_root / metadata.dir_to_str() / type_dirname).exists():
                metadata.AttemptNum += 1
        elif args.force or args.reckless:
            if not args.reckless:
                json_file = (
                    args.output_root
                    / metadata.dir_to_str()
                    / (metadata.prefix_to_str() + "_UnconvertedInfo.json")
                )
                if not json_file.exists():
                    raise ValueError(
                        "Unconverted info file (%s) does not exist for consistency checking. "
                        "Cannot use --force, use --reckless instead." % json_file
                    )
                json_obj = json.loads(json_file.read_text())
                if json_obj["Metadata"]["TMSMetaFileHash"] is not None:
                    if metadata.TMSMetaFileHash is None:
                        raise ValueError(
                            "Previous conversion did not use a TMS metadata file, "
                            "run with --reckless to ignore this error."
                        )
                    if json_obj["Metadata"]["TMSMetaFileHash"] != metadata.TMSMetaFileHash:
                        raise ValueError(
                            "TMS meta data file has changed since last conversion, "
                            "run with --reckless to ignore this error."
                        )
                elif (
                    json_obj["Metadata"]["TMSMetaFileHash"] is None
                    and metadata.TMSMetaFileHash is not None
                ):
                    raise ValueError(
                        "Previous conversion used a TMS metadata file, "
                        "run with --reckless to ignore this error."
                    )
                if hash_file_dir(args.source, False) != json_obj["InputHash"]:
                    raise ValueError(
                        "Source file(s) have changed since last conversion, "
                        "run with --reckless to ignore this error."
                    )
            shutil.rmtree(args.output_root / metadata.dir_to_str() / type_dirname)
            silentremove(args.output_root / metadata.dir_to_str() / "nii")
            for filepath in (args.output_root / metadata.dir_to_str()).glob("*.json"):
                silentremove(filepath)
        else:
            raise RuntimeError(
                "Output directory exists, run with --force to remove outputs and re-run."
            )

    manual_arg = {
        "MagneticFieldStrength": args.field_strength,
        "InstitutionName": args.institution,
    }

    run_conversion(
        args.source,
        args.output_root,
        metadata,
        lut_file,
        args.verbose,
        args.parrec,
        False,
        linking,
        manual_arg,
        args.force_dicom,
        args.anonymize,
        args.date_shift_days,
        manual_names,
        None,
    )


def update(args: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("directory", type=Path, help="Existing RADIFOX Directory to update.")
    parser.add_argument("-l", "--lut-file", type=Path, help="Lookup table file.")
    parser.add_argument(
        "--force", action="store_true", help="Force run even if it would be skipped."
    )
    parser.add_argument("--verbose", action="store_true", help="Verbose output.")
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

    args = parser.parse_args(args)

    session_id = args.directory.name
    subj_id = args.directory.parent.name

    json_file = args.directory / "_".join([subj_id, session_id, "UnconvertedInfo.json"])
    if not json_file.exists():
        safe_json_file = args.directory / "_".join(
            [subj_id, "-".join(session_id.split("-")[:-1]), "UnconvertedInfo.json"]
        )
        if not safe_json_file.exists():
            raise ValueError("Unconverted info file (%s) does not exist." % json_file)
        json_file = safe_json_file
    json_obj = json.loads(json_file.read_text())

    metadata = Metadata.from_dict(json_obj["Metadata"])
    if session_id != metadata.SessionID:
        metadata.AttemptNum = int(session_id.split("-")[-1])
    # noinspection PyProtectedMember
    output_root = (
        Path(*args.directory.parts[:-2])
        if metadata._NoProjectSubdir
        else Path(*args.directory.parts[:-3])
    )

    if args.lut_file is None:
        # noinspection PyProtectedMember
        if metadata._NoProjectSubdir:
            lut_file = output_root / (metadata.projectname + "-lut.csv")
        else:
            lut_file = output_root / metadata.projectname / (metadata.projectname + "-lut.csv")
    else:
        lut_file = args.lut_file
    lookup_dict = (
        LookupTable(lut_file, metadata.ProjectID, metadata.SiteID).LookupDict
        if lut_file.exists()
        else {}
    )

    manual_json_file = args.directory / (metadata.prefix_to_str() + "_ManualNaming.json")
    manual_names = json.loads(manual_json_file.read_text()) if manual_json_file.exists() else {}

    if not args.force and (
        version_check(json_obj["__version__"]["radifox"], __version__)
        and json_obj["LookupTable"]["LookupDict"] == lookup_dict
        and json_obj["ManualNames"] == manual_names
    ):
        print(
            "No action required. Software version, LUT dictionary and naming dictionary match for %s."
            % args.directory
        )
        return

    parrec = (args.directory / "parrec").exists()
    type_dir = args.directory / ("%s" % "parrec" if parrec else "dcm")

    mkdir_p(args.directory / "prev")
    for filename in ["nii", "qa", json_file.name]:
        if (args.directory / filename).exists():
            (args.directory / filename).rename(args.directory / "prev" / filename)
    try:
        run_conversion(
            type_dir,
            output_root,
            metadata,
            lut_file,
            args.verbose,
            parrec,
            True,
            None,
            json_obj.get("ManualArgs", {}),
            False,
            0,
            manual_names,
            json_obj["InputHash"],
        )
    except ExecError:
        logging.info("Exception caught during update. Resetting to previous state.")
        for filename in ["nii", "qa", json_file.name]:
            silentremove(args.directory / filename)
            if (args.directory / "prev" / filename).exists():
                (args.directory / "prev" / filename).rename(args.directory / filename)
    else:
        for dirname in ["stage", "proc"]:
            if (args.directory / dirname).exists():
                (args.directory / dirname / "CHECK").touch()
    silentremove(args.directory / "prev")


# TODO: Add "rename" command to rename sessions



from codecs import BOM_UTF8
from collections.abc import Sequence
import csv
from datetime import datetime, date, time, timedelta
import hashlib
import logging
from pathlib import Path
import re
import shutil
from subprocess import check_output

import nibabel as nib
from pydicom.dataset import Dataset, FileDataset, Tag

ORIENT_CODES = {"sagittal": "PIL", "coronal": "LIP", "axial": "LPS"}


# http://stackoverflow.com/a/22718321
def mkdir_p(path: Path, mode: int = 0o777) -> None:
    path.mkdir(mode=mode, parents=True, exist_ok=True)


def copytree_link(source: Path, dest: Path, symlink: bool) -> None:
    dest.mkdir(parents=True, exist_ok=True)
    for path in source.glob("*"):
        if path.is_file():
            if symlink:
                (dest / path.name).symlink_to(path)
            else:
                (dest / path.name).hardlink_to(path)
        elif path.is_dir():
            (dest / path.name).mkdir()
            copytree_link(path, dest / path.name, symlink)


# http://stackoverflow.com/a/10840586
def silentremove(filename: Path) -> None:
    import shutil
    import errno

    try:
        if filename.is_file():
            filename.unlink()
        else:
            shutil.rmtree(filename)
    except OSError as e:  # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred


def read_csv(csv_filename: Path) -> dict[str, list[str]]:
    data = csv_filename.read_bytes()
    codec = "UTF-8-SIG" if data.startswith(BOM_UTF8) else "UTF-8"
    data = data.decode(codec)
    line_sep = "\r\n" if "\r\n" in data else "\n"
    data = data.split(line_sep)
    out_dict = {key: [] for key in data[0].split(",")}
    for row in csv.DictReader(data):
        for key in out_dict.keys():
            out_dict[key].append(row[key])
    return out_dict


def convert_dicom_date(date_str: str) -> date:
    return datetime.strptime(date_str.replace("-", ""), "%Y%m%d").date()


def convert_dicom_time(time_str: str) -> time:
    return datetime.strptime(time_str.split(".")[0].ljust(6, "0"), "%H%M%S").time()


def convert_dicom_datetime(dt_str: str) -> date:
    return datetime.strptime(dt_str.replace("-", "").split(".")[0], "%Y%m%d%H%M%S")


vr_corr = {
    "CS": str,
    "DA": convert_dicom_date,
    "DS": float,
    "DT": convert_dicom_datetime,
    "FD": float,
    "FL": float,
    "IS": int,
    "LO": str,
    "LT": str,
    "OB": bytes,
    "PN": str,
    "SH": str,
    "SL": int,
    "SS": int,
    "ST": str,
    "TM": convert_dicom_time,
    "UI": str,
    "UL": int,
    "US": int,
}


def extract_de(
    ds: Dataset | FileDataset,
    label: str,
    series_uid,
    keep_list: bool = False,
) -> None | tuple | float | int | str:
    if label not in ds:
        return tuple() if keep_list else None
    de = ds[label]
    if de.VM == 0:
        return tuple() if keep_list else None
    value = [de.value] if de.VM == 1 else de.value
    try:
        out_list = []
        for item in value:
            out_list.append(vr_corr[de.VR](item))
        out_list = tuple(out_list)
    except ValueError:
        logging.warning(
            "Data element (%s) will not conform to required type (%s) for %s."
            % (de.name, str(vr_corr[de.VR]), series_uid)
        )
        return tuple() if keep_list else None
    return out_list[0] if (len(out_list) == 1 and not keep_list) else out_list


def is_intstr(test_str: str) -> bool:
    try:
        int(test_str)
    except ValueError:
        return False
    return True


def reorient(input_file: Path, orientation: str) -> bool:
    # noinspection PyTypeChecker
    input_obj = nib.Nifti1Image.load(input_file)
    input_orient = "".join(nib.aff2axcodes(input_obj.affine))
    target_orient = ORIENT_CODES[orientation]
    logging.debug("Reorienting %s from %s to %s." % (input_file, input_orient, target_orient))
    if input_orient != tuple(target_orient):
        try:
            orig_ornt = nib.orientations.io_orientation(input_obj.affine)
            targ_ornt = nib.orientations.axcodes2ornt(target_orient)
            ornt_xfm = nib.orientations.ornt_transform(orig_ornt, targ_ornt)

            affine = input_obj.affine.dot(nib.orientations.inv_ornt_aff(ornt_xfm, input_obj.shape))
            data = nib.orientations.apply_orientation(input_obj.dataobj, ornt_xfm)
            nib.Nifti1Image(data, affine, input_obj.header).to_filename(str(input_file))
            return True
        except ValueError:
            logging.warning("Reorientation failed for %s" % input_file)
            return False


def allowed_archives() -> (list[str], list[str]):
    allowed_exts = []
    allowed_names = []
    for names, extensions, _ in shutil.get_unpack_formats():
        allowed_exts.extend(extensions)
        allowed_names.append(names)
    if ".zip" in allowed_exts:
        allowed_exts.append(".zip.zip")
    return allowed_names, allowed_exts


def extract_archive(input_zipfile: Path, output_dir: Path) -> None:
    logging.info("Extracting archive")
    mkdir_p(output_dir)
    # noinspection PyTypeChecker
    shutil.unpack_archive(input_zipfile, output_dir)
    logging.info("Extraction complete")


def make_tuple(item: bytes | str | Sequence) -> tuple:
    if isinstance(item, (bytes, str)):
        return tuple([item])
    return tuple(item) if isinstance(item, Sequence) else tuple([item])


def remove_created_files(filename: Path) -> None:
    for imgname in [
        f
        for f in filename.parent.glob(filename.name + "*")
        if re.search(filename.name + r"_*[A-Za-z0-9_]*\..+$", f.name)
    ]:
        imgname.unlink()


def parse_dcm2niix_filenames(stdout: str) -> list[Path]:
    filenames = []
    for line in stdout.split("\n"):
        if line.startswith("Convert "):  # output
            fname = Path(str(re.search(r"\S+/\S+", line).group(0)))
            filenames.append(fname.resolve())
    return filenames


def add_acq_num(name: str, count: int) -> str:
    prefix = "_".join(name.split("_")[:-1])
    contrast_arr = name.split("_")[-1].split("-")
    addons = "" if len(contrast_arr) == 6 else ("-" + "-".join(contrast_arr[6:]))
    base_contrast = "-".join(contrast_arr[:6])
    return prefix + "_" + base_contrast + ("-ACQ%d" % count) + addons


FILE_OCTAL = 0o660
DIR_OCTAL = 0o2770


def has_permissions(path: Path, octal: int = DIR_OCTAL) -> bool:
    return int("0o" + oct(path.stat().st_mode)[-4:], 8) == octal


def recursive_chmod(
    directory: Path,
    dir_octal: int = DIR_OCTAL,
    file_octal: int = FILE_OCTAL,
) -> None:
    if not directory.exists():
        return
    elif directory.is_file():
        directory.chmod(file_octal)
    else:
        directory.chmod(dir_octal)
        for item in directory.rglob("*"):
            if item.is_dir():
                item.chmod(dir_octal)
            elif item.is_file():
                item.chmod(file_octal)


def find_closest(target: int, to_check: list[int]) -> int | None:
    if len(to_check) < 1:
        return None
    elif len(to_check) == 1:
        return to_check[0]
    signed_dists = []
    for check_val in to_check:
        signed_dists.append((check_val - target, check_val))
    min_dist = min([abs(val[0]) for val in signed_dists])
    candidates = [val[1] for val in signed_dists if abs(val[0]) == min_dist]
    return candidates[0] if len(candidates) == 1 else min(candidates)


def hash_file(
    filename: Path,
    include_names: bool = True,
    hashfunc: str = "sha256",
    *,
    _bufsize=2**18,
) -> str:
    hashobj = hashlib.new(hashfunc)
    if include_names:
        hashobj.update(filename.name.encode())
    with filename.open("rb") as fp:
        buf = bytearray(_bufsize)  # Reusable buffer to reduce allocations.
        view = memoryview(buf)
        while True:
            # noinspection PyUnresolvedReferences
            size = fp.readinto(buf)
            if size == 0:
                break  # EOF
            hashobj.update(view[:size])
    return str(hashobj.hexdigest())


def hash_dir(directory, include_names: bool = True, hashfunc: str = "sha256") -> str:
    hashobj = hashlib.new(hashfunc)
    for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
        hashobj.update(hash_file_dir(path, include_names, hashfunc).encode())
    if include_names:
        hashobj.update(directory.name.encode())
    return str(hashobj.hexdigest())


def hash_file_dir(file_dir: Path, include_names: bool = True, hashfunc: str = "sha256") -> str:
    if file_dir.is_file():
        return hash_file(file_dir, include_names=include_names, hashfunc=hashfunc)
    elif file_dir.is_dir():
        return hash_dir(file_dir, include_names=include_names, hashfunc=hashfunc)


def hash_file_list(
    file_list: list[Path], include_names: bool = True, hashfunc: str = "sha256"
) -> str:
    hashobj = hashlib.new(hashfunc)
    for path in file_list:
        hashobj.update(hash_file(path, include_names, hashfunc).encode())
    return str(hashobj.hexdigest())


def hash_value(value: str | None, hashfunc: str = "sha256") -> str | None:
    if value is None:
        return None
    m = hashlib.new(hashfunc)
    m.update(value.encode())
    return str(m.hexdigest())


def p_add(path: Path, extra: str) -> Path:
    return path.parent / (path.name + extra)


def get_software_versions() -> dict[str, str]:
    dcm2niix_version = (
        check_output("dcm2niix --version; exit 0", shell=True)
        .decode()
        .strip()
        .split("\n")[-1]
        .strip()
    )
    return {"dcm2niix": dcm2niix_version}


def version_check(saved_version: str, current_version: str) -> bool:
    if "dev" in saved_version or "dev" in current_version:
        return False
    saved_arr = saved_version.split("-")[0].split(".")
    current_arr = current_version.split("-")[0].split(".")
    for saved, current in zip(saved_arr, current_arr):
        if int(saved) < int(current):
            return False
    return True


def shift_date(datetime_str: str, date_shift_days: int = 0) -> str:
    orig_date = datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
    return (orig_date + timedelta(days=date_shift_days)).strftime("%Y-%m-%d %H:%M:%S")


def none_to_float(value: float | None) -> float:
    return float("inf") if value is None else value


def get_flattened_dataset(dataset: FileDataset | Dataset) -> Dataset:
    return Dataset({de.tag: de for de in dataset.iterall() if de.VR != "SQ"})


EXCLUDE_TAGS = [
    Tag("SharedFunctionalGroupsSequence"),
    Tag("PerFrameFunctionalGroupsSequence"),
    Tag("DimensionIndexSequence"),
    Tag("NumberOfFrames"),
    Tag("SourceImageEvidenceSequence"),
    Tag("ReferencedImageEvidenceSequence"),
    Tag("PixelData"),
]


def fix_sf_headers(dataset: Dataset) -> Dataset:
    if "EffectiveEchoTime" in dataset:
        dataset.EchoTime = dataset.EffectiveEchoTime
    scan_seq: list = (
        (
            dataset.ScanningSequence
            if dataset["ScanningSequence"].VM > 1
            else [dataset.ScanningSequence]
        )
        if "ScanningSequence" in dataset
        else []
    )
    if "EchoPulseSequence" in dataset:
        if dataset.EchoPulseSequence != "SPIN":
            scan_seq.append("GR")
        if dataset.EchoPulseSequence != "GRADIENT":
            scan_seq.append("SE")
    if dataset.get("InversionRecovery", "NO") == "YES":
        scan_seq.append("IR")
    if dataset.get("EchoPlanarPulseSequence", "NO") == "YES":
        scan_seq.append("EP")
    dataset.ScanningSequence = sorted(set(scan_seq))

    seq_var: list = (
        (
            dataset.SequenceVariant
            if dataset["SequenceVariant"].VM > 1
            else [dataset.SequenceVariant]
        )
        if "SequenceVariant" in dataset
        else []
    )
    if dataset.get("SegmentedKSpaceTraversal", "SINGLE") != "SINGLE":
        seq_var.append("SK")
    if dataset.get("MagnetizationTransfer", "NONE") != "NONE":
        seq_var.append("MTC")
    if dataset.get("SteadyStatePulseSequence", "NONE") != "NONE":
        seq_var.append("TRSS" if dataset.SteadyStatePulseSequence == "TIME_REVERSED" else "SS")
    if dataset.get("Spoiling", "NONE") != "NONE":
        seq_var.append("SP")
    if dataset.get("OversamplingPhase", "NONE") != "NONE":
        seq_var.append("OSP")
    if len(seq_var) == 0:
        seq_var.append("NONE")
    dataset.SequenceVariant = sorted(set(seq_var))

    scan_opts: list = (
        (dataset.ScanOptions if dataset["ScanOptions"].VM > 1 else [dataset.ScanOptions])
        if "ScanOptions" in dataset
        else []
    )
    if dataset.get("RectilinearPhaseEncodeReordering", "LINEAR") != "LINEAR":
        scan_opts.append("PER")
    frame_type3 = dataset.FrameType[2]
    if frame_type3 == "ANGIO":
        dataset.AngioFlag = "Y"
    if frame_type3.startswith("CARD"):
        scan_opts.append("CG")
    if frame_type3.endswith("RESP_GATED"):
        scan_opts.append("RG")
    if "PartialFourierDirection" in dataset:
        if dataset.PartialFourierDirection == "PHASE":
            scan_opts.append("PFP")
        elif dataset.PartialFourierDirection == "FREQUENCY":
            scan_opts.append("PFF")
    if dataset.get("SpatialPresaturation", "NONE") != "NONE":
        scan_opts.append("SP")
    if dataset.get("SpectrallySelectedSuppression", "NONE").startswith("FAT"):
        scan_opts.append("FS")
    if dataset.get("FlowCompensation", "NONE") != "NONE":
        scan_opts.append("FC")
    dataset.ScanOptions = sorted(set(scan_opts))
    return dataset


def create_sf_headers(dataset: Dataset | FileDataset) -> list[Dataset]:
    shared_ds = Dataset({de.tag: de for de in dataset if de.tag not in EXCLUDE_TAGS})
    shared_ds.file_meta = dataset.file_meta
    shared_ds.update(get_flattened_dataset(dataset.SharedFunctionalGroupsSequence[0]))
    flattened_frame_ds_list = [
        get_flattened_dataset(dataset.PerFrameFunctionalGroupsSequence[i])
        for i in range(len(dataset.PerFrameFunctionalGroupsSequence))
    ]

    sf_ds_list = []
    for flat_frame_ds in flattened_frame_ds_list:
        sf_ds = Dataset({de.tag: de for de in shared_ds})
        sf_ds.update(flat_frame_ds)
        sf_ds_list.append(fix_sf_headers(sf_ds))
    return sf_ds_list


def parse_dcm2niix_suffixes(filenames: list[Path], base: str, add_mag: bool) -> list[tuple[str]]:
    suffixes = [filename.name.replace(base, "") for filename in filenames]
    for i in range(len(suffixes)):
        suffixes[i] = suffixes[i].replace("_e", "_ECHO")
        suffixes[i] = suffixes[i].replace("_ph", "_PHA")
        suffixes[i] = suffixes[i].replace("_real", "_REA")
        suffixes[i] = suffixes[i].replace("_imaginary", "_IMA")
        suffixes[i] = suffixes[i].replace("_t", "_DYN")
    if add_mag or any([cplx in suffix for suffix in suffixes for cplx in ["PHA", "REA", "IMA"]]):
        for i in range(len(suffixes)):
            if not any([cplx in suffixes[i] for cplx in ["PHA", "REA", "IMA"]]):
                suffixes[i] += "_MAG"
    # Replace numbers after DYN with 1,2,3... in order of original number
    dyn_nums = sorted(
        {
            int(re.search(r"_DYN(\d+)", suffix).group(1))
            for suffix in suffixes
            if re.search(r"_DYN(\d+)", suffix) is not None
        }
    )
    for i in range(len(suffixes)):
        if re.search(r"_DYN(\d+)", suffixes[i]) is not None:
            dyn_idx = dyn_nums.index(int(re.search(r"_DYN(\d+)", suffixes[i]).group(1))) + 1
            suffixes[i] = re.sub(r"_DYN(\d+)", "_DYN%d" % dyn_idx, suffixes[i])
    # Do the same for echoes
    echo_nums = sorted(
        {
            int(re.search(r"_ECHO(\d+)", suffix).group(1))
            for suffix in suffixes
            if re.search(r"_ECHO(\d+)", suffix) is not None
        }
    )
    for i in range(len(suffixes)):
        if re.search(r"_ECHO(\d+)", suffixes[i]) is not None:
            dyn_idx = echo_nums.index(int(re.search(r"_ECHO(\d+)", suffixes[i]).group(1))) + 1
            suffixes[i] = re.sub(r"_ECHO(\d+)", "_ECHO%d" % dyn_idx, suffixes[i])
    return [tuple(sorted(suffix[1:].split("_"))) for suffix in suffixes]



#!/usr/bin/env python
from __future__ import annotations
from abc import ABC, abstractmethod
import argparse
from collections import defaultdict
import importlib.util
import inspect
import logging
from pathlib import Path

import nibabel as nib
import numpy as np
from .._version import __version__
from ..naming import ImageFile, ImageFilter, glob
from ..records import ProcessingModule

__all__ = ["Staging", "StagingPlugin"]


class Staging(ProcessingModule):
    name = "staging"
    version = __version__
    log_uses_filename = False
    skip_prov_write = ("session_target", "subject_target")

    @staticmethod
    def cli(args=None):
        parser = argparse.ArgumentParser()
        parser.add_argument("-s", "--subject-dir", type=Path, required=True)
        parser.add_argument("--image-types", type=str, nargs="+", required=True)
        parser.add_argument("--reg-filters", type=str, nargs="+", default=None)
        parser.add_argument("--keep-best-res", action="store_true", default=False)
        parser.add_argument("--plugin-paths", type=Path, nargs="+", default=None)
        parser.add_argument("--skip-default-plugins", action="store_true", default=False)
        parser.add_argument("--skip-set-sform", action="store_true", default=False)
        parsed = parser.parse_args(args)

        parsed.subject_dir = parsed.subject_dir.resolve()
        if not parsed.subject_dir.is_dir():
            parser.error(f"Subject directory ({parsed.subject_dir}) does not exist.")

        parsed.image_types = [
            ImageFilter.from_string(filter_str) for filter_str in parsed.image_types
        ]

        if parsed.plugin_paths is not None:
            parsed.plugin_paths = [p.resolve() for p in parsed.plugin_paths]
            for plugin_path in parsed.plugin_paths:
                if not plugin_path.is_file():
                    parser.error(f"Plugin file ({plugin_path}) does not exist.")

        if parsed.reg_filters is not None:
            parsed.reg_filters = [
                ImageFilter.from_string(filter_str) for filter_str in parsed.reg_filters
            ]
            for reg_filter in parsed.reg_filters:
                if reg_filter.acqdim is None:
                    parser.error(
                        f"Registration filters must provide an 'acqdim' filter."
                        f"{str(reg_filter)} was provided."
                    )

        session_imgs = {}
        for session in sorted(parsed.subject_dir.iterdir()):
            # Skip non-directories (if they exist)
            if not session.is_dir():
                continue
            # Get all images in session "nii" directory, sort by reverse name and skip "ND"
            all_imgs = glob(session / "nii" / "*.nii.gz")
            all_imgs = sorted(all_imgs, key=lambda x: x.name, reverse=True)
            all_imgs = [img for img in all_imgs if "ND" not in img.extras]
            if all_imgs:
                session_imgs[session] = all_imgs

        return {
            "session_filepaths": list(session_imgs.values()),
            "image_types": [parsed.image_types] * len(session_imgs),
            "keep_best_res": [parsed.keep_best_res] * len(session_imgs),
            "reg_filters": [parsed.reg_filters] * len(session_imgs),
            "plugin_paths": [parsed.plugin_paths] * len(session_imgs),
            "skip_default_plugins": [parsed.skip_default_plugins] * len(session_imgs),
            "skip_set_sform": [parsed.skip_set_sform] * len(session_imgs),
        }

    @staticmethod
    def run(
        session_filepaths: list[list[ImageFile]],
        image_types: list[list[ImageFilter]],
        keep_best_res: list[bool],
        plugin_paths: list[list[Path]],
        reg_filters: list[list[ImageFilter] | None],
        skip_default_plugins: list[bool],
        skip_set_sform: list[bool],
    ):
        # For each session, find images that match the contrast filters
        session_imgs = {}
        for (
            all_imgs,
            img_filters,
            best_res,
            proc_plugin_paths,
            skip_defaults,
            skip_set_sform_qform,
        ) in zip(
            session_filepaths,
            image_types,
            keep_best_res,
            plugin_paths,
            skip_default_plugins,
            skip_set_sform,
        ):
            if not all_imgs:
                continue
            # Create "stage" directory
            session = all_imgs[0].parent.parent
            (session / "stage").mkdir(exist_ok=True, parents=True)

            # Filter images by image filters
            filtered_imgs = []
            for img_filter in img_filters:
                # Get a list of images that match the filter (continue if none)
                imgs = img_filter.filter(all_imgs)
                if not imgs:
                    continue

                # Load plugins
                proc_plugins = []
                if proc_plugin_paths is not None:
                    for plugin_path in proc_plugin_paths:
                        proc_plugins.extend(load_plugins(plugin_path))
                if not skip_defaults:
                    proc_plugins.extend([MEMPRAGEPlugin, MP2RAGEPlugin])

                for plugin in proc_plugins:
                    plugin_imgs = plugin.filter(imgs)
                    other_imgs = [img for img in imgs if img not in plugin_imgs]
                    out_imgs = plugin.run(plugin_imgs)
                    imgs = out_imgs + other_imgs

                # Keep only the best resolution image from each contrast (if needed)
                if best_res:
                    # Get existing images in "stage" directory
                    stage_imgs = [img for img in imgs if img.parent.name == "stage"]
                    # If there is are 3D images, pick the one that is the "most isotropic"
                    imgs_3d = ImageFilter(acqdim="3D").filter(imgs)
                    if imgs_3d:
                        # Select the image with the lowest anisotropy
                        imgs = [pick_most_isotropic(imgs_3d)]
                    else:
                        # Sort images by slice spacing and/or slice thickness (select first)
                        imgs = [pick_smallest_slices(imgs)]
                    # Remove staged images that are not the best resolution
                    for img in stage_imgs:
                        if img not in imgs:
                            logging.warning(
                                f"Removing staged {img} because it is not the best image."
                            )
                            # img.path.unlink()
                # Add image(s) to filtered image list
                filtered_imgs.extend(imgs)
            if not skip_set_sform_qform:
                # Fix sform/qform of filtered images
                filtered_imgs = [fix_sform_qform(img) for img in filtered_imgs]

            # Remove staged images that are not in the filtered image list
            for stage_img in (session / "stage").iterdir():
                if stage_img.name not in [img.name for img in filtered_imgs]:
                    stage_img.unlink()

            session_imgs[session] = filtered_imgs

        reg_filters = reg_filters[0]
        if reg_filters is None:
            # No registration targets
            logging.info("No registration targets provided. Skipping.")
            session_targets = {session: None for session in session_imgs}
            subject_target = None
        else:
            # Subject targets are chosen over all sessions (first for tie)
            # Session targets are chosen over all images in each session
            # Ties in priority are resolved by resolution:
            # 1. For 3D images, most isotropic
            # 2. For 2D images, thinnest slice spacing
            # Find session targets
            session_targets: dict[Path, ImageFile] = {}
            for session, imgs in session_imgs.items():
                for reg_filter in reg_filters:
                    # Currenly use private _filter_dict to check acqdim
                    # Radifox should be updated to expose these parameters readonly
                    best_func = (
                        pick_most_isotropic if reg_filter.acqdim == "3D" else pick_smallest_slices
                    )
                    filtered = reg_filter.filter(imgs)
                    if filtered:
                        session_targets[session] = best_func(filtered)
                        break

            # Find subject target
            subject_target = None
            for img_filter in reg_filters:
                filtered = img_filter.filter(list(session_targets.values()))
                if filtered:
                    subject_target = filtered[0]
                    break
            if subject_target is None:
                raise ValueError("No target images found for subject.")

            # Symlink target images
            for session, img in session_targets.items():
                (session / "stage" / "session-target").symlink_to(
                    Path("..", img.path.relative_to(session))
                )
                (session / "stage" / "subject-target").symlink_to(
                    Path("..", "..", subject_target.path.relative_to(session.parent))
                )

            for session, imgs in session_imgs.items():
                logging.info("---")
                logging.info(session)
                logging.info("---")
                for img in imgs:
                    logging.info(f"{str(img.path)}")
            logging.info("---")
            logging.info("Registration Targets")
            logging.info("---")
            logging.info(subject_target.path)
            for session, img in session_targets.items():
                logging.info(f"{session}: {str(img.path)}")

        return [
            {
                "staged_files": imgs,
                "session_target": session_targets[session] if session in session_targets else None,
                "subject_target": subject_target if session in session_targets else None,
            }
            for session, imgs in session_imgs.items()
        ]


def fix_sform_qform(img: ImageFile) -> ImageFile:
    """Conform the qform/sform matrix of an image header."""
    out_fpath = img.parent.parent / "stage" / f"{img.stem}_hdrfix.nii.gz"
    obj = nib.Nifti1Image.load(img.path)
    obj.set_qform(obj.get_sform(), 2)
    obj.set_sform(obj.get_qform(), 1)
    obj.to_filename(out_fpath)
    return ImageFile(out_fpath)


def pick_most_isotropic(imgs: list[ImageFile]) -> ImageFile:
    """Pick the most isotropic image from a list of images."""
    # Return first image if only one image
    if len(imgs) == 1:
        return imgs[0]
    # Calculate aniontropy for each 3D image
    aniso = [
        np.abs(1 - (np.mean(img.info.AcquiredResolution) / img.info.SliceThickness)) for img in imgs
    ]
    # Select the image with the lowest anisotropy
    return imgs[np.argmin(aniso)]


def pick_smallest_slices(imgs: list[ImageFile]) -> ImageFile:
    """Pick the image with the smallest slice spacing from a list of images."""
    # Return first image if only one image
    if len(imgs) == 1:
        return imgs[0]
    # Get slice spacing for each image, sort, and pick first
    return sorted(
        imgs,
        key=lambda x: x.info.SliceThickness if x.info.SliceSpacing is None else x.info.SliceSpacing,
    )[0]


def load_plugins(plugin_path: Path) -> list[StagingPlugin]:
    """Load plugins from a plugin file."""
    # Get the module name from the plugin file name
    spec = importlib.util.spec_from_file_location(plugin_path.stem, plugin_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Inspect the module and find all subclasses of the specified base class
    return [
        member
        for name, member in inspect.getmembers(module, inspect.isclass)
        if issubclass(member, StagingPlugin) and member is not StagingPlugin
    ]


class StagingPlugin(ABC):
    @staticmethod
    @abstractmethod
    def filter(images: list[ImageFile]) -> list[ImageFile]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def run(images: list[ImageFile]) -> list[ImageFile]:
        raise NotImplementedError

    @staticmethod
    def sort_by_series(imgs: list[ImageFile]) -> list[list[ImageFile]]:
        """Sort images by series ID and return a list of images for each series ID."""
        series_dict = {}
        for img in imgs:
            if img.series_id not in series_dict:
                series_dict[img.series_id] = []
            series_dict[img.series_id].append(img)
        return list(series_dict.values())


class MEMPRAGEPlugin(StagingPlugin):
    @staticmethod
    def filter(images: list[ImageFile]) -> list[ImageFile]:
        return ImageFilter(
            modality="T1",
            technique="IRFSPGR",
            extras=lambda x: any("ECHO" in s or s == "SUM" for s in x),
        ).filter(images)

    @staticmethod
    def run(images: list[ImageFile]) -> list[ImageFile]:
        out_imgs = []
        for img_set in MEMPRAGEPlugin.sort_by_series(images):
            # Choose a SUM image if both echoes and SUM are available
            sum_imgs = [img for img in img_set if "SUM" in img.extras]
            if sum_imgs:
                out_imgs.append(sum_imgs[0])
            else:
                out_imgs.append(MEMPRAGEPlugin.sum_memprage(img_set))
        return out_imgs

    @staticmethod
    def sum_memprage(imgs: list[ImageFile]) -> ImageFile:
        """Create a sum image from a list of MEMPRAGE echo images."""
        temp_img = sorted(imgs, key=lambda x: x.name)[0]
        out_fpath = temp_img.path.parent.parent / "stage" / f"{temp_img.stem}_sum.nii.gz"
        obj = nib.load(temp_img.path)
        sum_data = np.sum(
            [nib.Nifti1Image.load(img.path).get_fdata(dtype=np.float32) for img in imgs], axis=0
        )
        nib.Nifti1Image(sum_data, None, obj.header).to_filename(out_fpath)
        return ImageFile(out_fpath)


class MP2RAGEPlugin(StagingPlugin):
    CMPLX_IMG_TYPES = ("MAG", "PHA", "REA", "IMA")

    @staticmethod
    def filter(images: list[ImageFile]) -> list[ImageFile]:
        return ImageFilter(
            modality="T1",
            technique="IRFSPGR",
            extras=lambda x: any("INV" in s for s in x),
        ).filter(images)

    @staticmethod
    def run(images: list[ImageFile]) -> list[ImageFile]:
        out_imgs = []
        for img_set in MP2RAGEPlugin.sort_by_series(images):
            out_imgs.append(MP2RAGEPlugin.create_uniden(img_set))
        return out_imgs

    @staticmethod
    def create_uniden(imgs: list[ImageFile], gamma: float = 1e9) -> ImageFile:
        """Create an UNIDEN (denoised uniform) image from the MP2RAGE complex component images."""
        img_dict = defaultdict(list)
        for img in imgs:
            inv = next(ex for ex in img.extras if "INV" in ex)
            cmplx_type = next(
                (ct for ct in MP2RAGEPlugin.CMPLX_IMG_TYPES if ct in img.extras), "MAG"
            )
            img_dict[f"{inv}-{cmplx_type}"].append(img)
        data = {
            key: nib.Nifti1Image.load(img_dict[key][0].path).get_fdata(dtype=np.float32)
            for key in img_dict
        }
        if all(len(img_dict[f"INV{i}-{comp}"]) == 1 for i in (1, 2) for comp in ["REA", "IMA"]):
            temp_img = img_dict["INV1-REA"][0]
            for inv in ["INV1", "INV2"]:
                data[f"{inv}-CPX"] = data[f"{inv}-REA"] + data[f"{inv}-IMA"] * 1j
        elif all(len(img_dict[f"INV{i}-{comp}"]) == 1 for i in (1, 2) for comp in ["MAG", "PHA"]):
            temp_img = img_dict["INV1-MAG"][0]
            for inv in ["INV1", "INV2"]:
                data[f"{inv}-PHA"] = (
                    (data["INV1-PHA"] - data["INV1-PHA"].min())
                    / (data["INV1-PHA"].max() - data["INV1-PHA"].min())
                    * (2 * np.pi)
                )
                data[f"{inv}-CPX"] = data[f"{inv}-MAG"] * np.exp(data[f"{inv}-PHA"] * 1j)
        else:
            raise ValueError("Cannot create uniden image from provided images.")
        uniden = np.real(
            ((np.conjugate(data["INV1-CPX"]) * data["INV2-CPX"]) - gamma)
            / (np.abs(data["INV1-CPX"]) ** 2 + np.abs(data["INV2-CPX"]) ** 2 + 2 * gamma)
        )
        uniden = np.clip(uniden, -0.5, 0.5) + 0.5

        obj = nib.Nifti1Image.load(temp_img.path)
        out_fpath = temp_img.parent.parent / "stage" / f"{temp_img.stem}_uniden.nii.gz"
        nib.Nifti1Image(uniden, None, obj.header).to_filename(out_fpath)
        return ImageFile(out_fpath)


if __name__ == "__main__":
    Staging()



from __future__ import annotations
import json
import logging
from pathlib import Path
import re
from typing import Optional

from .utils import hash_file_dir

META_TIME_CODES = {1: "00", 2: "06", 3: "12", 4: "24", 5: "36", 6: "48"}


class Metadata:
    def __init__(
        self,
        project_id: str,
        subject_id: str,
        session_id: str,
        site_id: Optional[str] = None,
        no_project_subdir: bool = False,
    ) -> None:
        self.ProjectID = project_id.upper()
        self.SubjectID = subject_id.upper()
        self.SessionID = session_id.upper()
        self.SiteID = site_id.upper() if site_id is not None else None
        self.AttemptNum = None
        self.TMSMetaFileHash = None
        self._RawMetaFileObj = None
        self._NoProjectSubdir = no_project_subdir

    @classmethod
    def from_tms_metadata(cls, metadata_file: Path, no_project_subdir: bool = False) -> Metadata:
        metadata_obj = json.loads(metadata_file.read_text())["metadataFieldsToValues"]
        if "patient_id" in metadata_obj:
            site_id, subject_id = metadata_obj["patient_id"].split("-")
        else:
            site_id, subject_id = metadata_obj["site_id"], "900"
        if subject_id == "900":
            session_id = "99"
        else:
            session_id = None
            for key in metadata_obj.keys():
                if re.match(r"mri_timepoint\(\d+\)", key):
                    tp_num = int(re.findall(r"\d+", key)[0])
                    session_id = str(83 + tp_num) if tp_num > 6 else META_TIME_CODES[tp_num]
                    break
        out_cls = cls(
            "treatms", subject_id, session_id, site_id, no_project_subdir=no_project_subdir
        )
        out_cls.TMSMetaFileHash = hash_file_dir(metadata_file)
        out_cls._RawMetaFileObj = {re.sub(r"\([0-9]*\)", "", k): v for k, v in metadata_obj.items()}
        return out_cls

    @classmethod
    def from_dict(cls, dict_obj: dict) -> Metadata:
        out_cls = cls(
            dict_obj["ProjectID"],
            dict_obj["SubjectID"],
            dict_obj["SessionID"],
            dict_obj["SiteID"],
            dict_obj["_NoProjectSubdir"],
        )
        if "TMSMetaFileHash" in dict_obj:
            out_cls.TMSMetaFileHash = dict_obj["TMSMetaFileHash"]
            out_cls._RawMetaFileObj = dict_obj["_RawMetaFileObj"]
        return out_cls

    def __repr_json__(self) -> dict:
        skip_keys = ["AttemptNum"]
        if self.TMSMetaFileHash is None:
            skip_keys += ["TMSMetaFileHash", "_RawMetaFileObj"]
        return {k: v for k, v in self.__dict__.items() if k not in skip_keys}

    @property
    def projectname(self) -> str:
        return self.ProjectID.lower()

    def check_metadata(self) -> None:
        if self._RawMetaFileObj is not None and self.SiteID != self._RawMetaFileObj["site_id"]:
            logging.warning(
                "Site ID (%s) does not match site portion of Subject ID (%s). "
                "Using %s as Site ID." % (self._RawMetaFileObj["site_id"], self.SiteID, self.SiteID)
            )

    def prefix_to_str(self) -> str:
        if self.SiteID is None:
            return f"{self.ProjectID}-{self.SubjectID}_{self.SessionID}"
        else:
            return f"{self.ProjectID}-{self.SiteID}-{self.SubjectID}_{self.SessionID}"

    def dir_to_str(self) -> Path:
        subject_str = (
            f"{self.ProjectID}-{self.SubjectID}"
            if self.SiteID is None
            else f"{self.ProjectID}-{self.SiteID}-{self.SubjectID}"
        )
        # noinspection PyStringFormat
        output_dir = Path(
            subject_str,
            self.SessionID + ("" if self.AttemptNum is None else ("-%d" % self.AttemptNum)),
        )
        if not self._NoProjectSubdir:
            output_dir = Path(self.projectname, output_dir)
        return output_dir



from pathlib import Path
from typing import Union, List, Optional

from .utils import read_csv, is_intstr, hash_value


class LookupTable:
    def __init__(self, lut_file: Path, project_id: str, site_id: str) -> None:
        filename = lut_file.resolve().expanduser()
        lut = read_csv(filename)
        if site_id is None:
            site_id = "NONE"
        self.LookupDict = {}
        for row, (project, site) in enumerate(zip(lut["Project"], lut["Site"])):
            site_name = "NONE" if site.upper() == "NONE" else site
            if is_intstr(site_name) and is_intstr(site_id):
                site_name = str(int(site_name))
                site_id = str(int(site_id))
            if site_name.lower() == site_id.lower() and project.lower() == project_id.lower():
                inst_name = (
                    "NONE"
                    if lut["InstitutionName"][row].upper() == "NONE"
                    else lut["InstitutionName"][row]
                )
                if inst_name not in self.LookupDict:
                    self.LookupDict[inst_name] = {}
                if lut["SeriesDescription"][row] in self.LookupDict[inst_name]:
                    raise ValueError(
                        "Series description (%s) already exists for site (%s) and institution name (%s)"
                        % (lut["SeriesDescription"][row], site_id, inst_name)
                    )
                self.LookupDict[inst_name][lut["SeriesDescription"][row]] = lut["OutputFilename"][
                    row
                ]

    def __repr_json__(self) -> dict:
        return self.__dict__

    def anonymize(self):
        old_inst_names = []
        for inst_name in self.LookupDict:
            if inst_name.upper() != "NONE":
                self.LookupDict[hash_value(inst_name)] = self.LookupDict[inst_name]
                old_inst_names.append(inst_name)
        for key in old_inst_names:
            del self.LookupDict[key]

    def check(self, inst_name: str, series_desc: str) -> Union[List[Optional[str]], bool]:
        # Deal with extras from PARRECs
        if series_desc.startswith("WIP "):
            series_desc = series_desc[4:]
        if series_desc.endswith(" CLEAR") or series_desc.endswith(" SENSE"):
            series_desc = series_desc[:-6]
        for item in [inst_name, "NONE"]:
            if item in self.LookupDict:
                if series_desc in self.LookupDict[item]:
                    if self.LookupDict[item][series_desc].upper() == "FALSE":
                        return False
                    lookup_arr = self.LookupDict[item][series_desc].split("-")
                    if len(lookup_arr) < 6:
                        lookup_arr += ["None"] * (6 - len(lookup_arr))
                    return [None if item.upper() == "NONE" else item for item in lookup_arr]
        return [None] * 6



import os
import itertools
import json
from collections import defaultdict
from pathlib import Path
import secrets

from flask import (
    Flask,
    request,
    render_template,
    jsonify,
    send_from_directory,
    url_for,
    session,
    redirect,
)
import yaml

from ..conversion.json import JSONObjectEncoder, NoIndent

DATA_DIR = Path(os.environ.get("QA_DATA_DIR", "/data")).resolve()
SECRET_KEY = os.environ.get("QA_SECRET_KEY", secrets.token_urlsafe())

app = Flask(__name__)
app.secret_key = secrets.token_hex()

with app.app_context():
    print(
        f"Access the QA Webapp Directly at: "
        f"http://{os.environ['QA_HOST']}:{os.environ['QA_PORT']}/login?key={SECRET_KEY}"
    )
    print(f"QA Data Directory: {DATA_DIR}")
    print(f"QA Secret Key: {SECRET_KEY}")


@app.before_request
def require_authentication():
    if session.get("qa_secret_key") != SECRET_KEY and request.endpoint != "login":
        return redirect(url_for("login"))


@app.route("/login", methods=["GET", "POST"])
def login():
    error = None
    if request.method == "POST":
        test_key = request.form["key"]
    else:
        test_key = request.args.get("key")

    if test_key is not None:
        if test_key == SECRET_KEY:
            session["qa_secret_key"] = test_key
            return redirect(url_for("index"))
        else:
            error = "Invalid Key"
    return render_template("login.html", error=error)


@app.route("/")
def index():
    projects = sorted([proj.name for proj in DATA_DIR.glob("*") if proj.is_dir()])
    return render_template("index.html", projects=projects)


@app.route("/<project_id>/")
def project(project_id):
    proj_dir = DATA_DIR / project_id
    subjects = sorted(
        [
            pat.name
            for pat in proj_dir.glob("*")
            if pat.is_dir()
            if pat.is_dir() and not pat.name.startswith(".")
        ]
    )
    return render_template("project.html", project_id=project_id, subjects=subjects)


@app.route("/<project_id>/<subject_id>/")
def subject(project_id, subject_id):
    pat_dir = DATA_DIR / project_id / subject_id
    sessions = sorted(
        [
            session_path.name
            for session_path in pat_dir.glob("*")
            if session_path.is_dir() and not session_path.name.startswith(".")
        ]
    )
    return render_template(
        "subject.html", project_id=project_id, subject_id=subject_id, sessions=sessions
    )


@app.route("/set_mode/<mode>/")
def set_mode(mode):
    session["qa_mode"] = mode
    return redirect(request.referrer)


@app.route("/<project_id>/<subject_id>/<session_id>/")
def qa_page(project_id, subject_id, session_id):
    if session.get("qa_mode", "conversion") == "conversion":
        return conversion_qa(project_id, subject_id, session_id)
    else:
        return processing_qa(project_id, subject_id, session_id)


def conversion_qa(project_id, subject_id, session_id):
    # Get subject and session information
    session_dir = DATA_DIR / project_id / subject_id / session_id
    subjects = sorted(
        [
            pat.name
            for pat in (DATA_DIR / project_id).glob("*")
            if pat.is_dir() and not pat.name.startswith(".")
        ]
    )
    pat_idx = subjects.index(subject_id)
    prev_subject = subjects[pat_idx - 1] if pat_idx > 0 else None
    next_subject = subjects[pat_idx + 1] if pat_idx < (len(subjects) - 1) else None
    sessions = sorted(
        [
            session_path.name
            for session_path in (DATA_DIR / project_id / subject_id).glob("*")
            if session_path.is_dir() and not session_path.name.startswith(".")
        ]
    )
    session_idx = sessions.index(session_id)
    prev_session = sessions[session_idx - 1] if session_idx > 0 else None
    next_session = sessions[session_idx + 1] if session_idx < (len(sessions) - 1) else None

    # Get conversion information
    json_files = (session_dir / "nii").glob("*.json")
    manual_json_path = (
        DATA_DIR
        / project_id
        / subject_id
        / session_id
        / "_".join([subject_id, session_id, "ManualNaming.json"])
    )
    manual_json = json.loads(manual_json_path.read_text()) if manual_json_path.exists() else {}
    conversion_images = []
    for jsonfile in json_files:
        si = json.loads(jsonfile.read_text())["SeriesInfo"]
        created_by = (
            "MANUAL"
            if any([item is not None for item in si["ManualName"]])
            else ("LOOKUP" if any([item is not None for item in si["LookupName"]]) else "PREDICTED")
        )
        manual_name = ""
        if si["SourcePath"] in manual_json:
            if manual_json[si["SourcePath"]] is False:
                created_by = "IGNORED"
            else:
                image_arr = si["NiftiName"].split("_")[-1].split("-")[:-1]
                manual_name = "/".join(
                    [item if item is not None else "--" for item in manual_json[si["SourcePath"]]]
                )
                if any(
                    [
                        j is not None and i != j
                        for i, j in zip(image_arr, manual_json[si["SourcePath"]])
                    ]
                ):
                    created_by = "MODIFIED"
        study_num, series_num = si["NiftiName"].split("_")[-2].split("-")
        image_obj = {
            "acq_date_time": si["AcqDateTime"],
            "series_number": series_num,
            "study_number": study_num,
            "acq_number": "%02d" % int(si["SeriesUID"].split(".")[-1]),
            "image_name": si["NiftiName"].split("_")[-1],
            "created_by": created_by,
            "institution_name": si["InstitutionName"],
            "manufacturer": si["Manufacturer"].upper(),
            "model": si["ScannerModelName"],
            "field_strength": si["MagneticFieldStrength"],
            "series_description": si["SeriesDescription"],
            "sequence_type": si["SequenceType"],
            "sequence_name": si["SequenceName"],
            "sequence_variant": si["SequenceVariant"],
            "scan_options": si["ScanOptions"],
            "acquired_resolution": (
                si["AcquiredResolution"] if si["AcquiredResolution"] is not None else [0, 0]
            ),
            "field_of_view": si["FieldOfView"] if si["FieldOfView"] is not None else [0, 0],
            "slice_thickness": si["SliceThickness"],
            "slice_spacing": si["SliceSpacing"],
            "num_slices": si["NumFiles"],
            "echo_time": si["EchoTime"],
            "repetition_time": si["RepetitionTime"],
            "inversion_time": si["InversionTime"],
            "flip_angle": si["FlipAngle"],
            "echo_train_length": si["EchoTrainLength"],
            "contrast_agent": si["ExContrastAgent"],
            "study_description": si["StudyDescription"],
            "body_part": si["BodyPartExamined"],
            "source_path": si["SourcePath"],
            "manual_name": manual_name,
        }
        img_path = session_dir / "qa" / "conversion" / jsonfile.name.replace(".json", ".png")
        if img_path.exists():
            image_obj["image_src"] = img_path.name
        conversion_images.append(image_obj)

    return render_template(
        "conversion.html",
        conversion_images=sorted(
            conversion_images,
            key=lambda x: (x["study_number"], x["series_number"], x["acq_number"]),
        ),
        project_id=project_id,
        subject_id=subject_id,
        session_id=session_id,
        next_session=next_session,
        prev_session=prev_session,
        prev_subject=prev_subject,
        next_subject=next_subject,
    )


def processing_qa(project_id, subject_id, session_id):
    # Get subject and session information
    session_dir = DATA_DIR / project_id / subject_id / session_id
    subjects = sorted(
        [
            pat.name
            for pat in (DATA_DIR / project_id).glob("*")
            if pat.is_dir() and not pat.name.startswith(".")
        ]
    )
    pat_idx = subjects.index(subject_id)
    prev_subject = subjects[pat_idx - 1] if pat_idx > 0 else None
    next_subject = subjects[pat_idx + 1] if pat_idx < (len(subjects) - 1) else None
    sessions = sorted(
        [
            session_path.name
            for session_path in (DATA_DIR / project_id / subject_id).glob("*")
            if session_path.is_dir() and not session_path.name.startswith(".")
        ]
    )
    session_idx = sessions.index(session_id)
    prev_session = sessions[session_idx - 1] if session_idx > 0 else None
    next_session = sessions[session_idx + 1] if session_idx < (len(sessions) - 1) else None

    # Get processing information
    prov_files = list(
        itertools.chain(
            (session_dir / "stage").glob("*.prov"),
            (session_dir / "proc").glob("*.prov"),
        )
    )
    prov_objs = defaultdict(dict)
    for provfile in prov_files:
        prov_obj = yaml.safe_load(provfile.read_text())
        prov_objs[prov_obj["Module"]][prov_obj["Id"]] = prov_obj
    prov_objs = {
        k: v
        for k, v in sorted(
            prov_objs.items(), key=lambda x: min(val["StartTime"] for val in x[1].values())
        )
    }

    for module_str, provs in prov_objs.items():
        for idstr, prov_obj in provs.items():
            prov_obj["OutputQA"] = {}
            for key, val in prov_obj["Outputs"].items():
                prov_obj["OutputQA"][key] = {}
                if not isinstance(val, list):
                    val = [val]
                for v in val:
                    filestr = v.split(":")[0]
                    filepath = session_dir.parent.parent / filestr
                    qa_path = (
                        filepath.parent.parent
                        / "qa"
                        / module_str.split(":")[0]
                        / (filepath.name.replace(".nii.gz", ".png"))
                    )
                    prov_obj["OutputQA"][key][filestr] = (
                        (qa_path.parent.parent.parent.name, qa_path.parent.name, qa_path.name)
                        if qa_path.exists()
                        else None
                    )

    return render_template(
        "processing.html",
        processing_results=prov_objs,
        project_id=project_id,
        subject_id=subject_id,
        session_id=session_id,
        next_session=next_session,
        prev_session=prev_session,
        prev_subject=prev_subject,
        next_subject=next_subject,
    )


def save_manual(json_obj, filepath):
    for key in json_obj:
        json_obj[key] = NoIndent(json_obj[key])

    filepath.write_text(json.dumps(json_obj, indent=4, sort_keys=True, cls=JSONObjectEncoder))
    filepath.chmod(0o660)


def add_manual_entry(project_id, subject_id, session_id, key, value):
    filepath = (
        DATA_DIR
        / project_id
        / subject_id
        / session_id
        / "_".join([subject_id, session_id, "ManualNaming.json"])
    )
    json_obj = json.loads(filepath.read_text()) if filepath.exists() else {}
    json_obj[key] = value
    save_manual(json_obj, filepath)


def update_manual_entry(project_id, subject_id, session_id, data):
    filepath = (
        DATA_DIR
        / project_id
        / subject_id
        / session_id
        / "_".join([subject_id, session_id, "ManualNaming.json"])
    )
    json_obj = json.loads(filepath.read_text()) if filepath.exists() else {}

    name_list = [
        data.get(key, "").upper()
        for key in ["body_part", "modality", "technique", "acq_dim", "orient", "ex_contrast"]
    ]
    name_list = [None if item.strip() == "" else item for item in name_list]

    existing = [False] * 6
    if data["source"] in json_obj and json_obj[data["source"]] is not False:
        existing = [item is not None for item in json_obj[data["source"]]]

    name_list = [
        None if (new == old and not exist) else new
        for new, old, exist in zip(name_list, data["original_name"].split("-"), existing)
    ]

    if all([item is None for item in name_list]):
        del json_obj[data["source"]]
    else:
        json_obj[data["source"]] = name_list

    save_manual(json_obj, filepath)


@app.route("/manual-entry", methods=["POST"])
def manual_entry():
    data = request.form
    update_manual_entry(
        data["project"],
        data["subject"],
        data["session"],
        data,
    )
    return "Manual Name Added/Updated"


@app.route("/ignore-entry", methods=["POST"])
def ignore_entry():
    data = request.form
    add_manual_entry(
        data["project"],
        data["subject"],
        data["session"],
        data["source"],
        False,
    )
    return "Image Ignored"


@app.route("/ignore-btn", methods=["POST"])
def ignore_btn():
    data = request.get_json()
    add_manual_entry(
        data["project"],
        data["subject"],
        data["session"],
        data["source"],
        False,
    )
    return jsonify(message="Image Ignored")


@app.route("/change-btn", methods=["POST"])
def change_btn():
    data = request.get_json()
    update_manual_entry(
        data["project"],
        data["subject"],
        data["session"],
        data,
    )
    return jsonify(message="Body Part Updated")


@app.route("/image/<project_id>/<subject_id>/<session_id>/<qa_dir>/<image_name>")
def image(project_id, subject_id, session_id, qa_dir, image_name):
    return send_from_directory(
        str(DATA_DIR / project_id / subject_id / session_id / "qa" / qa_dir), image_name
    )



import logging
from collections import defaultdict
from pathlib import Path
import shutil
from typing import Optional

from pydicom import dcmread, Dataset, FileDataset
from pydicom.errors import InvalidDicomError
from pydicom.dicomdir import DicomDir

from .base import BaseInfo, BaseSet, ImageOrientation, TruncatedImageValue, MATCHING_ITEMS
from .lut import LookupTable
from .metadata import Metadata
from .utils import mkdir_p, extract_de, create_sf_headers

DCM_HEADER_ATTRS = [
    "InstitutionName",
    "Manufacturer",
    ("ManufacturerModelName", "ScannerModelName"),
    ("DeviceSerialNumber", "DeviceIdentifier"),
    "SeriesDescription",
    "MagneticFieldStrength",
    ("MRAcquisitionType", "AcquisitionDimension"),
    ("SpacingBetweenSlices", "SliceSpacing"),
    "SliceThickness",
    "FlipAngle",
    "RepetitionTime",
    "EchoTime",
    "InversionTime",
    "TriggerTime",
    "EchoTrainLength",
    "AcquisitionMatrix",
    ("ScanningSequence", "SequenceType"),
    "ScanOptions",
    "ImageType",
    "SeriesNumber",
    "ComplexImageComponent",
    "BodyPartExamined",
    "StudyDescription",
    "SequenceVariant",
    ("PixelSpacing", "ReconResolution"),
    "SoftwareVersions",
    "NumberOfAverages",
    "PercentSampling",
    "ReceiveCoilName",
    "PixelBandwidth",
    ("VariableFlipAngleFlag", "VariableFlipAngle"),
]

DCM_HEADER_LISTS = [
    "ScanningSequence",
    "SequenceVariant",
    "ImageType",
    "ScanOptions",
    "SoftwareVersions",
]


class DicomInfo(BaseInfo):
    def __init__(
        self,
        source_path: Path,
        ds: Dataset | FileDataset,
        series_uid: str,
        num_frames: int,
        multiframe: bool,
    ) -> None:
        super().__init__(source_path)
        self.SeriesUID = series_uid
        self.StudyUID = getattr(ds, "StudyInstanceUID", None)
        self.NumFiles = num_frames
        self.MultiFrame = multiframe
        for item in DCM_HEADER_ATTRS:
            get_item, set_item = item if isinstance(item, tuple) else (item, item)
            setattr(
                self,
                set_item,
                extract_de(ds, get_item, self.SeriesUID, get_item in DCM_HEADER_LISTS),
            )
        if self.SeriesDescription is None:
            self.SeriesDescription = ds.ProtocolName if hasattr(ds, "ProtocolName") else ""
        series_date = extract_de(ds, "SeriesDate", self.SeriesUID, False)
        series_time = extract_de(ds, "SeriesTime", self.SeriesUID, False)
        self.AcqDateTime = " ".join(
            [
                str(series_date) if series_date is not None else "0001-01-01",
                str(series_time) if series_date is not None else "00:00:00",
            ]
        )
        self.Manufacturer = (
            "" if self.Manufacturer is None else self.Manufacturer.upper().split(" ")[0]
        )
        if (0x2005, 0x1444) in ds:
            turbo = int(ds[(0x2005, 0x1444)].value)
            self.EchoTrainLength = turbo if turbo > 0 else self.EchoTrainLength
        if (0x2001, 0x1013) in ds:
            try:
                self.EPIFactor = int(ds[(0x2001, 0x1013)].value)
            except ValueError:
                pass
        if self.InversionTime == 0.0 and (0x2001, 0x101B) in ds:
            try:
                self.InversionTime = float(ds[(0x2001, 0x101B)].value)
            except ValueError:
                pass
            if self.TriggerTime is not None and "mp2rage" in self.SeriesDescription.lower():
                self.InversionTime = self.TriggerTime
                self.TriggerTime = None
        if self.AcquisitionMatrix is not None:
            # noinspection PyUnresolvedReferences
            self.AcquisitionMatrix = (
                [self.AcquisitionMatrix[0], self.AcquisitionMatrix[3]]
                if self.AcquisitionMatrix[1] == 0
                else [self.AcquisitionMatrix[2], self.AcquisitionMatrix[1]]
            )
            self.ReconMatrix = [getattr(ds, "Columns", 0), getattr(ds, "Rows", 0)]
            self.FieldOfView = [
                res * num for res, num in zip(self.ReconResolution, self.ReconMatrix)
            ]
            self.AcquiredResolution = [
                fov / num for fov, num in zip(self.FieldOfView, self.AcquisitionMatrix)
            ]
            self.FieldOfView = [round(fov, 2) for fov in self.FieldOfView]
            self.AcquiredResolution = [round(res, 5) for res in self.AcquiredResolution]
            self.ReconResolution = [round(res, 5) for res in self.ReconResolution]
        self.SequenceName = getattr(ds, "SequenceName", getattr(ds, "PulseSequenceName", None))
        if self.SequenceName is None and (0x0019, 0x109C) in ds:
            self.SequenceName = str(ds[(0x0019, 0x109C)].value)
        self.ExContrastAgent = getattr(
            ds, "ContrastBolusAgent", getattr(ds, "ContrastBolusAgentSequence", None)
        )
        self.ImageOrientationPatient = ImageOrientation(
            getattr(ds, "ImageOrientationPatient", None)
        )
        self.SliceOrientation = self.ImageOrientationPatient.get_plane()
        self.ImagePositionPatient = TruncatedImageValue(getattr(ds, "ImagePositionPatient", None))
        if self.ComplexImageComponent is None and any(
            [
                comp in self.ImageType
                for comp in ["M", "P", "R", "I", "MAGNITUDE", "PHASE", "REAL", "IMAGINARY"]
            ]
        ):
            if "M" in self.ImageType or "MAGNITUDE" in self.ImageType:
                self.ComplexImageComponent = "MAGNITUDE"
            elif "P" in self.ImageType or "PHASE" in self.ImageType:
                self.ComplexImageComponent = "PHASE"
            elif "R" in self.ImageType or "REAL" in self.ImageType:
                self.ComplexImageComponent = "REAL"
            elif "I" in self.ImageType or "IMAGINARY" in self.ImageType:
                self.ComplexImageComponent = "IMAGINARY"


class DicomSet(BaseSet):
    def __init__(
        self,
        source: Path,
        output_root: Path,
        metadata_obj: Metadata,
        lut_obj: LookupTable,
        remove_identifiers: bool = False,
        date_shift_days: int = 0,
        manual_names: Optional[dict] = None,
        input_hash: Optional[str] = None,
    ) -> None:
        super().__init__(
            source,
            output_root,
            metadata_obj,
            lut_obj,
            remove_identifiers,
            date_shift_days,
            manual_names,
            input_hash,
        )

        logging.info("Loading DICOMs.")
        for dcmdir in sorted((output_root / self.Metadata.dir_to_str() / "dcm").glob("*")):
            ds = dcmread(str(sorted(dcmdir.glob("*"))[0]), stop_before_pixels=True)
            if ds.SOPClassUID == "1.2.840.10008.5.1.4.1.1.4.1":
                sfds_dict = defaultdict(list)
                for sfds in sorted(create_sf_headers(ds), key=lambda x: x.InstanceNumber):
                    sfds_dict[get_intra_series_meta(sfds)].append(sfds)
                for i, sfds_list in enumerate(sfds_dict.values()):
                    self.SeriesList.append(
                        DicomInfo(
                            dcmdir,
                            sfds_list[0],
                            sfds_list[0].SeriesInstanceUID + (".%02d" % i),
                            len(sfds_list),
                            True,
                        )
                    )
            else:
                self.SeriesList.append(
                    DicomInfo(dcmdir, ds, dcmdir.name, len(list(dcmdir.glob("*"))), False)
                )

        study_nums, series_nums = self.get_unique_study_series(self.SeriesList)
        for di in self.SeriesList:
            logging.info("Processing %s" % di.SourcePath)
            if di.should_convert():
                di.create_image_name(
                    self.Metadata.prefix_to_str(),
                    study_nums[di.SourcePath],
                    series_nums[di.SourcePath],
                    self.LookupTable,
                    self.ManualNames,
                )

        logging.info("Generating unique names")
        self.generate_unique_names()


def get_intra_series_meta(ds: Dataset) -> tuple:
    return tuple(
        ImageOrientation(getattr(ds, item, None))
        if item == "ImageOrientationPatient"
        else extract_de(ds, item, ds.SeriesInstanceUID)
        for item in MATCHING_ITEMS
    )


def sort_dicoms(dcm_dir: Path, force_dicom: bool = False) -> None:
    logging.info("Sorting DICOMs")
    valid_dcms = []
    for filepath in dcm_dir.rglob("*"):
        if filepath.is_file():
            try:
                ds = dcmread(str(filepath), stop_before_pixels=True)
            except (InvalidDicomError, KeyError):
                if force_dicom:
                    try:
                        ds = dcmread(str(filepath), stop_before_pixels=True, force=True)
                        ds.save_as(str(filepath), write_like_original=False)
                    except ValueError:
                        continue
                else:
                    continue
            if isinstance(ds, DicomDir):
                continue
            if ds.SOPClassUID not in ["1.2.840.10008.5.1.4.1.1.4", "1.2.840.10008.5.1.4.1.1.4.1"]:
                continue
            if getattr(ds, "SeriesInstanceUID", None) is None:
                continue
            if ds.SOPClassUID == "1.2.840.10008.5.1.4.1.1.4.1":
                valid_dcms.extend((filepath, sf) for sf in create_sf_headers(ds))
            else:
                valid_dcms.append((filepath, ds))
    if not valid_dcms:
        raise ValueError("No valid DICOMs found in %s, try re-running with --force-dicom" % dcm_dir)

    # Sort valid DICOMs according to series and intra-series metadata
    valid_dcms = sorted(valid_dcms, key=lambda x: (x[1].SeriesInstanceUID, x[1].InstanceNumber))
    dicom_sort_dict = {}
    for dcm_file, dcm_ds in valid_dcms:
        uid = dcm_ds.SeriesInstanceUID
        intra_series = get_intra_series_meta(dcm_ds)
        if (uid, intra_series) not in dicom_sort_dict:
            dicom_sort_dict[(uid, intra_series)] = []
        dicom_sort_dict[(uid, intra_series)].append((dcm_file, dcm_ds))

    # Get new series (based on series/intra-series metadata)
    new_series_uids = {}
    count_dict = defaultdict(int)
    for scan in dicom_sort_dict:
        count_dict[scan[0]] += 1
        new_series_uids[scan] = scan[0] + (".%02d" % count_dict[scan[0]])

    # Find new series that have scans that share DICOM files (multi-frame)
    uids_by_files = defaultdict(list)
    for scan, dcm_list in dicom_sort_dict.items():
        uids_by_files[tuple(dcm[0] for dcm in dcm_list)].append(scan)
    # Merge new series uids that share DICOM files
    for scans in uids_by_files.values():
        if len(scans) > 1:
            for scan in scans:
                new_series_uids[scan] = ".".join(new_series_uids[scan].split(".")[:-1])

    for new_dcm_dir in new_series_uids.values():
        mkdir_p(dcm_dir / new_dcm_dir)
    inst_nums = defaultdict(lambda: defaultdict(list))
    moved = set()
    for scan, dcm_tuples in dicom_sort_dict.items():
        output_dir = dcm_dir / new_series_uids[scan]
        for dcm_file, dcm_ds in dcm_tuples:
            if dcm_file in moved:
                continue
            dcm_file.rename(output_dir / dcm_file.name)
            moved.add(dcm_file)
            inst_nums[new_series_uids[scan]][dcm_ds.InstanceNumber].append(
                output_dir / dcm_file.name
            )

    new_dirs = list(new_series_uids.values())
    for item in dcm_dir.glob("*"):
        if item.name not in new_dirs:
            if item.is_dir():
                shutil.rmtree(item)
            elif item.is_file():
                item.unlink()

    logging.info("Checking for duplicate DICOM files")
    for uid in inst_nums:
        if any([len(inst_nums[uid][num]) > 1 for num in inst_nums[uid]]):
            logging.info("Possible duplicates found for %s" % uid)
            remove_duplicates(dcm_dir / uid)

    logging.info("Sorting complete")


def remove_duplicates(dcmdir: Path) -> None:
    inst_nums = defaultdict(list)
    for dcmfile in dcmdir.glob("*"):
        ds = dcmread(str(dcmfile), stop_before_pixels=True)
        inst_nums[ds.InstanceNumber].append((dcmfile, ds))
    count = 0
    for num in inst_nums:
        inst_nums[num] = sorted(
            inst_nums[num], key=lambda x: getattr(x[1], "InstanceCreationTime", 0)
        )
        for i in range(len(inst_nums[num]) - 1):
            for j in range(i + 1, len(inst_nums[num])):
                diff_keys = []
                dcmfile1, ds1 = inst_nums[num][i]
                dcmfile2, ds2 = inst_nums[num][j]
                for key in ds1.keys():
                    if key in [(0x0008, 0x0013), (0x0008, 0x0018)]:
                        continue
                    if (key not in ds2.keys()) or (ds1[key].value != ds2[key].value):
                        diff_keys.append(key)
                if len(diff_keys) == 0:
                    if dcmread(str(dcmfile1)).PixelData == dcmread(str(dcmfile2)).PixelData:
                        logging.debug("Found duplicate of %s" % dcmfile1)
                        logging.debug("Removing duplicate file %s" % dcmfile2)
                        dcmfile2.unlink()
                        count += 1
                        break
    if count > 0:
        logging.info("Removed %d duplicate DICOM files" % count)
    else:
        logging.info("No duplicate DICOM files found")



import logging
from pathlib import Path
import shutil
from typing import Optional

from .._version import __version__
from .dicom import DicomSet, sort_dicoms
from .logging import create_loggers
from .metadata import Metadata
from .lut import LookupTable
from .parrec import ParrecSet, sort_parrecs
from .utils import (
    mkdir_p,
    extract_archive,
    allowed_archives,
    copytree_link,
    DIR_OCTAL,
    has_permissions,
    hash_value,
)


class ExecError(Exception):
    pass


def run_conversion(
    source: Optional[Path],
    output_root: Path,
    metadata: Metadata,
    lut_file: Path,
    verbose: bool,
    parrec: bool,
    rerun: bool,
    link: Optional[str],
    manual_args: dict,
    force_dicom: bool,
    remove_identifiers: bool,
    date_shift_days: int,
    manual_names: dict,
    input_hash: Optional[str] = None,
) -> None:
    session_path = output_root / metadata.dir_to_str()
    mkdir_p(session_path)
    if not has_permissions(session_path, DIR_OCTAL):
        session_path.chmod(DIR_OCTAL)
    if not has_permissions(session_path.parent, DIR_OCTAL):
        session_path.parent.chmod(mode=DIR_OCTAL)

    create_loggers(output_root / metadata.dir_to_str() / "logs", "conversion", verbose)
    metadata.check_metadata()

    try:
        logging.info("Beginning scan conversion using RADIFOX v" + __version__)
        if remove_identifiers:
            logging.info(
                "Anonymization will be performed, including removal of copied source folders."
            )
        if not lut_file.exists():
            logging.warning("LUT file does not exist. Creating a blank file at %s." % lut_file)
            lut_file.write_text("Project,Site,InstitutionName,SeriesDescription,OutputFilename\n")
        lut = LookupTable(lut_file, metadata.ProjectID, metadata.SiteID)
        if metadata.AttemptNum is not None:
            logging.info("Multiple attempts found. This will be attempt #%d" % metadata.AttemptNum)
        if parrec:
            inst_name = (
                hash_value(manual_args["InstitutionName"])
                if remove_identifiers
                else manual_args["InstitutionName"]
            )
            logging.info(
                "PARREC source indicated. Using InstitutionName=%s and MagneticFieldStrength=%d"
                % (inst_name, manual_args["MagneticFieldStrength"])
            )
        logging.info("RADIFOX conversion starting: %s" % metadata.dir_to_str())
        type_folder = session_path / ("parrec" if parrec else "dcm")
        if not rerun:
            if source.is_dir():
                if link is not None:
                    logging.info("Linking files from source to %s folder" % type_folder.name)
                    if link not in ["symlink", "hardlink"]:
                        raise ValueError("Unsupported linking type.")
                    copytree_link(source, type_folder, link == "symlink")
                    logging.info("Linking complete")
                else:
                    logging.info("Copying files from source to %s folder" % type_folder.name)
                    # noinspection PyTypeChecker
                    shutil.copytree(source, type_folder, copy_function=shutil.copyfile)
                    logging.info("Copying complete")
            elif any([source.name.endswith(ext) for ext in allowed_archives()[1]]):
                extract_archive(source, type_folder)
            else:
                raise ValueError(
                    "Source is not a directory, but does not match one of "
                    "the available archive formats (%s)" % ", ".join(allowed_archives()[0])
                )
            if parrec:
                sort_parrecs(type_folder)
            else:
                sort_dicoms(type_folder, force_dicom)

        if parrec:
            img_set = ParrecSet(
                source,
                output_root,
                metadata,
                lut,
                remove_identifiers,
                date_shift_days,
                manual_names,
                input_hash=input_hash,
                manual_args=manual_args,
            )
        else:
            img_set = DicomSet(
                source,
                output_root,
                metadata,
                lut,
                remove_identifiers,
                date_shift_days,
                manual_names,
                input_hash=input_hash,
            )
        img_set.create_all_nii()
        img_set.generate_unconverted_info()

        if remove_identifiers:
            shutil.rmtree(type_folder)
        logging.info("RADIFOX conversion finished: %s" % metadata.dir_to_str())
    except KeyboardInterrupt:
        raise
    except:
        logging.exception("Fatal error occurred.")
        raise ExecError()



import datetime
import fcntl


def safe_append_to_file(filename, data):
    with open(filename, 'a') as file:
        fcntl.flock(file.fileno(), fcntl.LOCK_EX)  # Exclusive lock
        file.write(data)
        fcntl.flock(file.fileno(), fcntl.LOCK_UN)  # Release lock


def format_timedelta(delta: datetime.timedelta) -> str:
    # Extracting days and seconds
    days = delta.days
    seconds = delta.seconds

    # Converting seconds to hours, minutes, and seconds
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60

    # return formatted string
    return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"
